﻿//---------------------------------------------------------------------------
//
//   Spring
//
//          Copyright(C) 1999-2012, tetraface Inc.
//
//     Sample of Create plug-in for creating a spring.
//     Created DLL must be installed in "Plugins\Create" directory.
//
//    　バネを作成するクリエイトプラグインのサンプル。
//    　作成したDLLは"Plugins\Create"フォルダに入れる必要がある。
//
//
//---------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "MQPlugin.h"
#include "resource.h"

HINSTANCE hInstance;

BOOL CreateSpring(MQDocument doc);
BOOL CALLBACK SpringDialogProc(HWND, UINT, WPARAM, LPARAM);


//---------------------------------------------------------------------------
//  DllMain
//---------------------------------------------------------------------------
BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
	//ダイアログボックスの表示に必要なので、インスタンスを保存しておく
	hInstance = (HINSTANCE)hModule;

	//プラグインとしては特に必要な処理はないので、何もせずにTRUEを返す
    return TRUE;
}

//---------------------------------------------------------------------------
//  MQGetPlugInID
//    プラグインIDを返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT void MQGetPlugInID(DWORD *Product, DWORD *ID)
{
	// プロダクト名(制作者名)とIDを、全部で64bitの値として返す
	// 値は他と重複しないようなランダムなもので良い
	*Product = 0x56A31D20;
	*ID      = 0x70AE0C31;
}

//---------------------------------------------------------------------------
//  MQGetPlugInName
//    プラグイン名を返す。
//    この関数は[プラグインについて]表示時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT const char *MQGetPlugInName(void)
{
	// プラグイン名
	return "Create spring           Copyright(C) 1999-2012, tetraface Inc.";
}

//---------------------------------------------------------------------------
//  MQGetPlugInType
//    プラグインのタイプを返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT int MQGetPlugInType(void)
{
	// 作成用プラグインである
	return MQPLUGIN_TYPE_CREATE;
}

//---------------------------------------------------------------------------
//  MQEnumString
//    ポップアップメニューに表示される文字列を返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT const char *MQEnumString(int index)
{
	switch(index){
	case 0: return "Create spring";
	}
	return NULL;
}

//---------------------------------------------------------------------------
//  MQCreate
//    メニューから選択されたときに呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT BOOL MQCreate(int index, MQDocument doc)
{
	switch(index){
	case 0: return CreateSpring(doc);
	}
	return FALSE;
}

struct {
	float length;
	float rotate;
	float radius;
	int seg_u;
	int seg_v;
	float width;
} spring;

// バネを作成
BOOL CreateSpring(MQDocument doc)
{
	int u,v;
	int mat_index;

	// ダイアログボックスを表示
	if(DialogBox(hInstance, MAKEINTRESOURCE(DLG_SPRING), 
		MQ_GetWindowHandle(), SpringDialogProc) != TRUE)
		return FALSE;

	// バネ用のオブジェクトを作成
	MQObject obj = MQ_CreateObject();

	// 頂点を作成
	for(u=0; u<=spring.seg_u; u++)
	{
		for(v=0; v<spring.seg_v; v++)
		{
			MQPoint p;
			float u_angle = (float)u / spring.seg_u * spring.rotate * 2 * 3.14159f;
			float v_angle = (float)v / spring.seg_v * 2 * 3.14159f;

			p.y = (float)spring.length * u / spring.seg_u + (float)cos(v_angle) * spring.width;
			p.x = (float)cos(u_angle) * (spring.radius + (float)sin(v_angle) * spring.width);
			p.z = (float)sin(u_angle) * (spring.radius + (float)sin(v_angle) * spring.width);
			obj->AddVertex(p);
		}
	}

	// カレントマテリアルの番号を得る
	mat_index = doc->GetCurrentMaterialIndex();

	// 面を作成
	for(u=0; u<spring.seg_u; u++)
	{
		for(v=0; v<spring.seg_v; v++)
		{
			int index[4];
			index[0] = (u  )*spring.seg_v + v;
			index[1] = (u  )*spring.seg_v + (v+1)%spring.seg_v;
			index[2] = (u+1)*spring.seg_v + (v+1)%spring.seg_v;
			index[3] = (u+1)*spring.seg_v + v;
			// 面を追加
			int face_index = obj->AddFace(4, index);
			// 追加した面にカレントマテリアルを割り当て
			obj->SetFaceMaterial(face_index, mat_index);
		}
	}

	// オブジェクトをドキュメントへ登録
	doc->AddObject(obj);

	return TRUE;
}

// ダイアログボックスのコールバック関数
BOOL CALLBACK SpringDialogProc(
  HWND hwndDlg, // handle to dialog box
  UINT uMsg, // message
  WPARAM wParam, // first message parameter
  LPARAM lParam // second message parameter
)
{
	switch(uMsg){
	case WM_COMMAND:
		switch(LOWORD(wParam)){
		case IDOK:
			{
				char buf[64];

				if(GetDlgItemText(hwndDlg, EDIT_LENGTH, buf, sizeof(buf)) == 0) break;
				spring.length = (float)atof(buf);

				if(GetDlgItemText(hwndDlg, EDIT_ROTATE, buf, sizeof(buf)) == 0) break;
				spring.rotate = (float)atof(buf);

				if(GetDlgItemText(hwndDlg, EDIT_RADIUS, buf, sizeof(buf)) == 0) break;
				spring.radius = (float)atof(buf);

				if(GetDlgItemText(hwndDlg, EDIT_SEG_U, buf, sizeof(buf)) == 0) break;
				spring.seg_u = atoi(buf);

				if(GetDlgItemText(hwndDlg, EDIT_SEG_V, buf, sizeof(buf)) == 0) break;
				spring.seg_v = atoi(buf);

				if(GetDlgItemText(hwndDlg, EDIT_WIDTH, buf, sizeof(buf)) == 0) break;
				spring.width = (float)atof(buf);
			}
			EndDialog(hwndDlg, TRUE);
			break;
		case IDCANCEL:
			PostMessage(hwndDlg, WM_CLOSE, 0, 0);
			break;
		}
		break;

	case WM_INITDIALOG:
		SetDlgItemText(hwndDlg, EDIT_LENGTH, "100");
		SetDlgItemText(hwndDlg, EDIT_ROTATE, "8");
		SetDlgItemText(hwndDlg, EDIT_RADIUS, "20");
		SetDlgItemText(hwndDlg, EDIT_SEG_U, "100");
		SetDlgItemText(hwndDlg, EDIT_SEG_V, "10");
		SetDlgItemText(hwndDlg, EDIT_WIDTH, "4");
		return TRUE;

	case WM_CLOSE:
		EndDialog(hwndDlg, FALSE);
		break;
	}

	return FALSE;
}

 
