﻿#ifndef _STATION_SPY_PLUGIN_H_
#define _STATION_SPY_PLUGIN_H_

#include "MQBasePlugin.h"

class StationSpyPlugin : public MQStationPlugin
{
	friend class CMainDlg;

protected:
	CMainDlg *dlgMain;
	bool m_bActivate;
	bool m_bOnDraw;
	bool m_bOnUpdateScene;
	bool m_bOnUpdateUndo;

public:
	StationSpyPlugin();
	virtual ~StationSpyPlugin();

	// プラグインIDを返す。
	virtual void GetPlugInID(DWORD *Product, DWORD *ID);
	// プラグイン名を返す。
	virtual const char *GetPlugInName(void);
	// ポップアップメニューに表示される文字列を返す。
	virtual const char *EnumString(void);

	// アプリケーションの初期化
	virtual BOOL Initialize();
	// アプリケーションの終了
	virtual void Exit();

	// 表示・非表示切り替え要求
	virtual BOOL Activate(MQDocument doc, BOOL flag);
	// 表示・非表示状態の返答
	virtual BOOL IsActivated(MQDocument doc);
	// ウインドウの最小化への返答
	virtual void OnMinimize(MQDocument doc, BOOL flag);
	// プラグイン独自のメッセージの受け取り
	virtual int OnReceiveUserMessage(MQDocument doc, DWORD src_product, DWORD src_id, const char *description, void *message);
	// 描画時の処理
	virtual void OnDraw(MQDocument doc, MQScene scene, int width, int height);

	// ドキュメント初期化時
	virtual void OnNewDocument(MQDocument doc, const char *filename, NEW_DOCUMENT_PARAM& param);
	// ドキュメント終了時
	virtual void OnEndDocument(MQDocument doc);
	// ドキュメント保存時
	virtual void OnSaveDocument(MQDocument doc, const char *filename, SAVE_DOCUMENT_PARAM& param);
	// アンドゥ実行時
	virtual BOOL OnUndo(MQDocument doc, int undo_state);
	// リドゥ実行時
	virtual BOOL OnRedo(MQDocument doc, int redo_state);
	// アンドゥ状態更新時
	virtual void OnUpdateUndo(MQDocument doc, int undo_state, int undo_size);
	// オブジェクトの編集時
	virtual void OnObjectModified(MQDocument doc);
	// オブジェクトの選択状態の変更時
	virtual void OnObjectSelected(MQDocument doc);
	// カレントオブジェクトの変更時
	virtual void OnUpdateObjectList(MQDocument doc);
	// マテリアルのパラメータ変更時
	virtual void OnMaterialModified(MQDocument doc);
	// カレントマテリアルの変更時
	virtual void OnUpdateMaterialList(MQDocument doc);
	// シーン情報の変更時
	virtual void OnUpdateScene(MQDocument doc, MQScene scene);


	typedef bool (StationSpyPlugin::*ExecuteCallbackProc)(MQDocument doc);

	void Execute(ExecuteCallbackProc proc);

	// 面の削除
	bool DeleteFace(MQDocument doc);

protected:
	struct CallbackInfo {
		ExecuteCallbackProc proc;
	};

	// コールバックに対する実装部
	virtual bool ExecuteCallback(MQDocument doc, void *option);
};


#endif _STATION_SPY_PLUGIN_H_
