﻿//---------------------------------------------------------------------------
//
//   TTFont
//
//          Copyright(C) 1999-2012, tetraface Inc.
//
//     Sample of Create plug-in for creating a string with True-Type font.
//     This is just a sample. The precision of the triangulation is not 
//    enough.
//     Created DLL must be installed in "Plugins\Create" directory.
//
//    　True-Type Fontから文字列を作成するクリエイトプラグインのサンプル。
//    サンプルといってもソースコードは汚いので、参考にはならないかもしれ
//    ない。三角形分割処理は精度が良くなくてたまにおかしくなるし、速度の
//    最適化も出来てない代物なので、とりあえず生成出来るという程度でしか
//    なく、はっきり言って出来のいいものではない。
//    　作成したDLLは"Plugins\Create"フォルダに入れる必要がある。
//
//
//---------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "MQPlugin.h"
#include "Triangulate.h"
#include "resource.h"


HINSTANCE hInstance;

BOOL CreateFont(MQDocument doc);


//---------------------------------------------------------------------------
//  DllMain
//---------------------------------------------------------------------------
BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
	//ダイアログボックスの表示に必要なので、インスタンスを保存しておく
	hInstance = (HINSTANCE)hModule;

	//プラグインとしては特に必要な処理はないので、何もせずにTRUEを返す
    return TRUE;
}

//---------------------------------------------------------------------------
//  MQGetPlugInID
//    プラグインIDを返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT void MQGetPlugInID(DWORD *Product, DWORD *ID)
{
	// プロダクト名(制作者名)とIDを、全部で64bitの値として返す
	// 値は他と重複しないようなランダムなもので良い
	*Product = 0x56A31D20;
	*ID      = 0x5EAC34E0;
}

//---------------------------------------------------------------------------
//  MQGetPlugInName
//    プラグイン名を返す。
//    この関数は[プラグインについて]表示時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT const char *MQGetPlugInName(void)
{
	// プラグイン名
	return "Create Text with TrueType Font      Copyright(C) 1999-2012, tetraface Inc.";
}

//---------------------------------------------------------------------------
//  MQGetPlugInType
//    プラグインのタイプを返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT int MQGetPlugInType(void)
{
	// 作成用プラグインである
	return MQPLUGIN_TYPE_CREATE;
}

//---------------------------------------------------------------------------
//  MQEnumString
//    ポップアップメニューに表示される文字列を返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT const char *MQEnumString(int index)
{
	switch(index){
	case 0: return "Create text";
	}
	return NULL;
}

//---------------------------------------------------------------------------
//  MQCreate
//    メニューから選択されたときに呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT BOOL MQCreate(int index, MQDocument doc)
{
	switch(index){
	case 0: return CreateFont(doc);
	}
	return FALSE;
}


typedef enum SPTYPE {SPEdge, SPSpline} SPTYPE;
typedef struct _SPLINEPOINT {
	POINTFX p;
	SPTYPE type;
} SPLINEPOINT;

TTPoint PointFxToPoint(POINTFX f)
{
	TTPoint p;
	p.x = (float)f.x.fract / 65536.f + (float)f.x.value;
	p.y = (float)f.y.fract / 65536.f + (float)f.y.value;
	return p;
};


class MOutline
{
public:
	TTPolygon **Polygons;
	int Count;

	MOutline() {
		Polygons = NULL;
		Count = 0;
	}
	virtual ~MOutline(){
		Clear();
	}

	void Clear(void){
		if(Polygons) {
			for(int i=0; i<Count; i++)
				delete Polygons[i];
			free(Polygons);
			Polygons = NULL;
			Count = 0;
		}
	}

	int Add(TTPolygon *NewPoly)
	{
		Polygons = (TTPolygon **)realloc(Polygons, sizeof(TTPolygon *) * (Count+1));
		Polygons[Count++] = NewPoly;
		return Count; //return count of faces
	}
	int Delete(int index)
	{
		if(index < Count){
			delete Polygons[index];
			for(int i=index; i<Count-1; i++){
				Polygons[i] = Polygons[i+1];
			}
			Count--;
		}
		return Count;
	}

	bool GetOutline(HDC hdc, UINT uChar, int segment);
	int CheckCrossingLine(TTPoint l1, TTPoint l2);
	int CheckCrossingLine(TTPolygon *poly1, int index1, TTPolygon *poly2, int index2);
	bool Analyze(MQObject obj, float offset_x);
	bool IsInside(int out_index, int in_index);
};

bool MOutline::GetOutline(HDC hdc, UINT uChar, int segment)
{
	int ObjCount=0;
	GLYPHMETRICS gm;
	TEXTMETRIC tm;
	MAT2 mat2 = {{0,1}, {0,0}, {0,0}, {0,1}};

	//get size of memory for outline
	GetGlyphOutline(hdc, uChar, GGO_METRICS, &gm, 0, NULL, &mat2);
	DWORD BufSize = GetGlyphOutline(hdc, uChar, GGO_NATIVE, &gm, 0, NULL, &mat2);
	if(BufSize == 0 || BufSize == GDI_ERROR)
		return false;

	//alloc buffer
	void *Buffer = malloc(BufSize);
	if(Buffer == NULL)
		return false;

	//get outline
	if(GetGlyphOutline(hdc, uChar, GGO_NATIVE, &gm, BufSize, Buffer, &mat2) == GDI_ERROR)
	{
		free(Buffer);
		return false;
	}

	//get metrics
	float descent = 0;
	if(GetTextMetrics(hdc, &tm))
		descent = (float)tm.tmDescent / tm.tmHeight;

	int LastSize = BufSize;
	void *Ptr = Buffer;
	int PtfxCount;

	SPLINEPOINT *Ptfx = (SPLINEPOINT *)malloc(sizeof(SPLINEPOINT) * 1024);

	//analyze outline
	for(; LastSize>0; ObjCount++)
	{
		TTPOLYGONHEADER *Ttph = (TTPOLYGONHEADER *)Ptr;
		LastSize -= Ttph->cb;
		Ptr = ((char*)Ptr + Ttph->cb);

		if(Ttph->dwType != TT_POLYGON_TYPE)
			continue;
		int LastPoly = Ttph->cb - sizeof(TTPOLYGONHEADER);
		Ptfx[0].p = Ttph->pfxStart;
		Ptfx[0].type = SPEdge;
		PtfxCount = 1;

		//create poly-face
		TTPolygon *Poly = new TTPolygon;

		//get index
		TTPOLYCURVE *Ttpc = (TTPOLYCURVE *)&Ttph[1];
		for(;LastPoly>0;)
		{
			int pc = Ttpc->cpfx;
			LastPoly -= (pc-1)*sizeof(POINTFX)+sizeof(TTPOLYCURVE);
			POINTFX *pfx = Ttpc->apfx;
			switch(Ttpc->wType){
			case TT_PRIM_LINE:{
				for(int i=0; i<pc; i++,pfx++){
					Ptfx[i+PtfxCount].p = *pfx;
					Ptfx[i+PtfxCount].type = SPEdge;
				}
				PtfxCount += pc;
				break;}
			case TT_PRIM_QSPLINE:{
				for(int i=0; i<pc; i++,pfx++){
					Ptfx[PtfxCount+i].p = *pfx;
					Ptfx[PtfxCount+i].type = SPSpline;
				}
				PtfxCount += pc;
				Ptfx[PtfxCount-1].type = SPEdge;
				break;}
			}
			Ttpc = (TTPOLYCURVE *)pfx;
		}
		if(memcmp(&Ptfx[0].p, &Ptfx[PtfxCount-1].p, sizeof(POINTFX)) == 0)
			PtfxCount--;
		Ptfx[PtfxCount].p = Ptfx[0].p;
		Ptfx[PtfxCount].type = SPEdge;

		//set points
		for(int i=0; i<PtfxCount; i++)
		{
			TTPoint pp, pa, pb, pc;

			switch(Ptfx[i].type){
			case SPEdge:
				pp = PointFxToPoint(Ptfx[i].p);
				pp.x = pp.x / 2048.f;
				pp.y = (pp.y + descent) / 2048.f;
				Poly->Add(pp);
				break;
			case SPSpline:
				pb = PointFxToPoint(Ptfx[i].p);
				pa = PointFxToPoint(Ptfx[i-1].p);
				if(Ptfx[i-1].type != SPEdge) 
					pa = (pa + pb) / 2;
				pc = PointFxToPoint(Ptfx[i+1].p);
				if(Ptfx[i+1].type != SPEdge) 
					pc = (pc + pb) / 2;

				for(int s=1; s<segment; s++)
				{
					float t = (float)s / (float)segment;
					pp = (pa-2*pb+pc)*t*t + 2*(pb-pa)*t + pa;
					pp.x = pp.x / 2048.f;
					pp.y = (pp.y + descent) / 2048.f;
					Poly->Add(pp);
				}
				break;
			}
		}

		Add(Poly);
	}

	free(Ptfx);
	free(Buffer);

	return true;
}

int MOutline::CheckCrossingLine(TTPoint l1, TTPoint l2)
{
	for(int i=0; i<Count; i++)
	{
		TTPolygon *poly = Polygons[i];
		for(int j=0; j<poly->ptscount; j++)
		{
			TTPoint p1 = poly->pts[j];
			TTPoint p2 = poly->pts[j+1<poly->ptscount ? j+1 : 0];
			float n1 = (p1.x-l1.x) * (p2.x-l1.y) - (p1.y-l1.y) * (p2.x-l1.x);
			float n2 = (p1.x-l2.x) * (p2.x-l2.y) - (p1.y-l2.y) * (p2.x-l2.x);
			if(n1 * n2 < 0)
				return i;
		}
	}

	return -1;
}

int MOutline::CheckCrossingLine(TTPolygon *poly1, int index1, TTPolygon *poly2, int index2)
{
	TTPoint l1 = poly1->pts[poly1->indices->index[index1]];
	TTPoint l2 = poly2->pts[poly2->indices->index[index2]];
	for(int i=0; i<Count; i++)
	{
		TTPolygon *poly = Polygons[i];
		int icount = poly->indices->count;
		int *index = poly->indices->index;
		for(int j=0; j<icount; j++)
		{
			TTPoint p1 = poly->pts[index[j]];
			TTPoint p2 = poly->pts[index[j+1<icount ? j+1 : 0]];
			float n1 = (p1.x-l1.x) * (p2.y-l1.y) - (p1.y-l1.y) * (p2.x-l1.x);
			float n2 = (p1.x-l2.x) * (p2.y-l2.y) - (p1.y-l2.y) * (p2.x-l2.x);
			float n3 = (p1.x-l1.x) * (p1.y-l2.y) - (p1.y-l1.y) * (p1.x-l2.x);
			float n4 = (p2.x-l1.x) * (p2.y-l2.y) - (p2.y-l1.y) * (p2.x-l2.x);
			if(n1 * n2 <= 0 && n3 * n4<=0) {
				if(poly == poly1 && (j == index1 || (j+1)%icount == index1))
					continue;
				if(poly == poly2 && (j == index2 || (j+1)%icount == index2))
					continue;
				return i;
			}
		}
	}

	return -1;
}

bool MOutline::IsInside(int out_index, int in_index)
{
	if(out_index == in_index)
		return false;

	TTPolygon *outside, *inside;
	int outcount;
	TTPoint *op, ip;
	int i;

	outside = Polygons[out_index];
	outcount = outside->indices->count;

	// copy points to buffer
	op = (TTPoint *)malloc(sizeof(TTPoint) * (outcount+2));
	for(i=0; i<outcount; i++)
		op[i] = outside->pts[outside->indices->index[i]];
	op[outside->indices->count]   = op[0];
	op[outside->indices->count+1] = op[1];

	inside = Polygons[in_index];
	ip = inside->pts[0]; // check only one point

	int flag=0; //cross count  inside(1) or outside(0)
	int endv = outcount;
	char *lchk = (char *)malloc(sizeof(char) * outcount);
	memset(lchk, 0, sizeof(char) * outcount);

	for(i=0; i<outcount; i++)
	{
		float v = (ip.y-op[i].y) * (ip.y-op[i+1].y);
		if(v < 0) {
			float t = (ip.y - op[i].y) / (op[i+1].y - op[i].y);
			if(ip.x < op[i].x*(1-t) + op[i+1].x*t) {
				lchk[i] = 1;
				flag++;
			} else {
				lchk[i] = -1;
			}
		} else if(v > 0)
			lchk[i] = -1;
	}

	for(i=0; i<outcount; i++)
	{
		if(lchk[i] != 0)
			continue;

		if(ip.y != op[i].y)
			continue;

		int v1,v2;
		for(v1=i; v1!=v1-outcount; v1--) {
			if(ip.y != op[(v1+outcount)%outcount].y)
				break;
		}
		for(v2=i; v2!=v2+outcount; v2++) {
			if(ip.y != op[(v2+outcount)%outcount].y)
				break;
		}
		float v = (ip.y-op[(v1+outcount)%outcount].y) * (ip.y-op[v2%outcount].y);
		if(v < 0 && ip.x<op[(v1+1)%outcount].x) {
			for(; v1!=i; v1++)
				lchk[(v1+outcount)%outcount] = 1;
			for(--v2; v2!=i; v2--)
				lchk[v2%outcount] = 1;
			lchk[i] = 1;
			flag++;
		} else {
			for(; v1!=i; v1++)
				lchk[(v1+outcount)%outcount] = -1;
			for(--v2; v2!=i; v2--)
				lchk[v2%outcount] = -1;
			lchk[i] = -1;
		}
	}

	free(lchk);

	return (flag & 1);
}

bool MOutline::Analyze(MQObject obj, float offset_x)
{
	int i,j,k,l,m;

	// align clockwise
	for(i=0; i<Count; i++)
	{
		TTPolygon *poly = Polygons[i];
		if(poly->ptscount >= 3) {
			if(!poly->IsClockwise())
				poly->Invert();
		}
	}

	// initialize indices of points
	for(i=0; i<Count; i++)
		Polygons[i]->InitIndex();

	// piled outlines must be separated, but not implement

	// analyze inside
START_ANALYZE:
	for(i=0; i<Count; i++)
	{
#if 1
		// if this polygon is in other polygon, ignore
		for(j=0; j<Count; j++)
			if(IsInside(j, i))
				break;
		if(j < Count)
			continue;
#endif

		// check if other polygon is inside
		for(j=0; j<Count; j++)
		{
			if(!IsInside(i, j))
				continue;

#if 1
			TTPolygon *outside = Polygons[i];
			TTPolygon *inside = Polygons[j];
			int opc = outside->indices->count;
			int ipc = inside->indices->count;
			// search uncrossing line
			for(k=0; k<opc; k++)
			{
				for(l=0; l<ipc; l++)
				{
					if(CheckCrossingLine(outside, k, inside, l) < 0)
					{
						// merge points
						int *nvi = (int *)malloc(sizeof(int) * inside->ptscount);
						for(m=0; m<inside->ptscount; m++)
							nvi[m] = outside->Add(inside->pts[m]);

						// merge outlines
						int n=1;
						for(m=l; m>=0; m--,n++)
							outside->indices->Insert(k+n, nvi[inside->indices->index[m]]);
						for(m=ipc-1; m>=l; m--,n++)
							outside->indices->Insert(k+n, nvi[inside->indices->index[m]]);
						outside->indices->Insert(k+n, outside->indices->index[k]);

						free(nvi);
						Delete(j);
						goto START_ANALYZE;
					}
				}
			}
#else
			Delete(j);
			goto START_ANALYZE;
#endif
		}
	}

	// register polygons to object
	for(i=0; i<Count; i++)
	{
		TTPolygon *poly = Polygons[i];
		if(poly->ptscount < 2)
			continue;

		// add vertices
		int *vi = (int *)malloc(sizeof(int) * poly->ptscount);
		for(j=0; j<poly->ptscount; j++)
			vi[j] = obj->AddVertex( MQPoint(poly->pts[j].x + offset_x, poly->pts[j].y, 0.f) );

		// triangulate
		poly->Triangulate();
		for(int i=0; i<poly->triangle->count; i++)
		{
			Indices *indices = poly->triangle->index[i];
			int fvi[3];
			fvi[0] = vi[indices->index[0]];
			fvi[1] = vi[indices->index[2]];
			fvi[2] = vi[indices->index[1]];
			obj->AddFace(3, fvi);
		}

		free(vi);
	}

	return true;
}

bool CreateFont(MQObject obj, HDC hdc, wchar_t *wstr, int segment)
{
	int len = wcslen(wstr);
	if(len == 0)
		return false;

	// set object from font
	float offset_x = 0;
	for(int i=0; i<len; i++)
	{
		//convert unicode character to multi-byte
		char mchar[4];
		memset(mchar, 0, sizeof(mchar));
		WideCharToMultiByte(CP_OEMCP, 0, &wstr[i], 
			1, mchar, sizeof(mchar), NULL, NULL);

		//convert multi-byte to glyph
		UINT gchar=0;
		for(int i=0; i<4 && mchar[i]!=0; i++){
			gchar<<=8;
			unsigned char c = mchar[i];
			gchar |= (UINT)c;
		}

		// get outline
		MOutline outline;
		if(!outline.GetOutline(hdc, gchar, segment))
			continue;

		outline.Analyze(obj, offset_x);

		// get width and update offset
		ABC Abc;
		GetCharABCWidths(hdc, gchar, gchar, &Abc);
		offset_x += (float)(Abc.abcA + Abc.abcB + Abc.abcC) / 2048.f;
	}

	//move to client area


	return true;
}

MQObject CreateTextObject(const char *text, const char *font, int segment)
{
	MQObject obj = MQ_CreateObject();

	// create DC for drawing font
	HDC hdc = CreateDC("DISPLAY",NULL,NULL,NULL);

	// set info of font
	LOGFONT lf;
	memset(&lf, 0, sizeof(LOGFONT));
	lf.lfHeight = 2048;
	lf.lfWidth  = 0;
	lf.lfEscapement = 0;
	lf.lfWeight = FW_NORMAL;
	lf.lfItalic = 0;
	lf.lfUnderline = 0;
	lf.lfStrikeOut = 0;
	lf.lfCharSet   = DEFAULT_CHARSET;
	lf.lfOutPrecision = OUT_DEFAULT_PRECIS; //OUT_STROKE_PRECIS;
	lf.lfClipPrecision = CLIP_DEFAULT_PRECIS;
	lf.lfQuality = DEFAULT_QUALITY; //PROOF_QUALITY;
	lf.lfPitchAndFamily = FF_DONTCARE | DEFAULT_PITCH;
	strncpy_s(lf.lfFaceName, sizeof(lf.lfFaceName), font, LF_FACESIZE-1);

	// create font
	HFONT hfont = CreateFontIndirect(&lf);
	HFONT hfontold = (HFONT)SelectObject(hdc, hfont);

	// set unicode text
	int wsize = MultiByteToWideChar(CP_OEMCP, MB_PRECOMPOSED, text, -1, NULL, 0);
	if(wsize != 0)
	{
		wchar_t *wch = (wchar_t *)malloc(wsize * sizeof(wchar_t));
		MultiByteToWideChar(CP_OEMCP, MB_PRECOMPOSED, text, -1, wch, wsize);

		CreateFont(obj, hdc, wch, segment);

		free(wch);
	}

	// delete objects
	SelectObject(hdc, hfontold);
	DeleteObject(hfont);
	DeleteDC(hdc);

	return obj;
}

class FontData {
public:
	int segment;
	char *text;
	char font[LF_FACESIZE];

	FontData() {
		text = NULL;
	}
	virtual ~FontData() {
		if(text != NULL)
			free(text);
	}
	void AllocText(int len) {
		if(text != NULL)
			free(text);
		text = (char *)malloc(len+1);
	}
} font;

// フォント列挙のコールバック関数
int CALLBACK FontEnumProc(LPLOGFONT lpLogFont, LPTEXTMETRIC lpTextMetric, DWORD dwType, LPARAM lpData)
{
	if(dwType == TRUETYPE_FONTTYPE)
	{
		SendDlgItemMessage((HWND)lpData, COMBO_FONT, CB_ADDSTRING, 0, (LPARAM)lpLogFont->lfFaceName);
		SendDlgItemMessage((HWND)lpData, COMBO_FONT, CB_SETCURSEL, (WPARAM)0, 0);
	}

	return 1; // continue to enumeration
}

// ダイアログボックスのコールバック関数
BOOL CALLBACK FontDialogProc(
  HWND hwndDlg, // handle to dialog box
  UINT uMsg, // message
  WPARAM wParam, // first message parameter
  LPARAM lParam // second message parameter
)
{
	HDC hdc;

	switch(uMsg){
	case WM_COMMAND:
		switch(LOWORD(wParam)){
		case IDOK:
			{
				char buf[64];
				int len, index;

				if(GetDlgItemText(hwndDlg, EDIT_SEGMENT, buf, sizeof(buf)) == 0) break;
				font.segment = atoi(buf);
				
				len = SendDlgItemMessage(hwndDlg, EDIT_TEXT, WM_GETTEXTLENGTH, 0, 0);
				font.AllocText(len+1);
				SendDlgItemMessage(hwndDlg, 
					EDIT_TEXT, WM_GETTEXT, (WPARAM)len+1, (LPARAM)font.text);

				index = SendDlgItemMessage(hwndDlg, COMBO_FONT, CB_GETCURSEL, 0, 0);
				if(index != CB_ERR)
					SendDlgItemMessage(hwndDlg, 
						COMBO_FONT, CB_GETLBTEXT, (WPARAM)index, (LPARAM)font.font);
			}
			EndDialog(hwndDlg, TRUE);
			break;
		case IDCANCEL:
			PostMessage(hwndDlg, WM_CLOSE, 0, 0);
			break;
		}
		break;

	case WM_INITDIALOG:
		SetDlgItemText(hwndDlg, EDIT_SEGMENT, "5");

	    SendDlgItemMessage(hwndDlg, COMBO_FONT, CB_RESETCONTENT, 0, 0 );
		hdc = GetDC(hwndDlg);
		EnumFonts(hdc, NULL, (FONTENUMPROC)FontEnumProc, (LPARAM)hwndDlg);
		ReleaseDC(hwndDlg, hdc);
		SendDlgItemMessage(hwndDlg, COMBO_FONT, CB_SETCURSEL, (WPARAM)0, 0);

 		return TRUE;

	case WM_CLOSE:
		EndDialog(hwndDlg, FALSE);
		break;
	}

	return FALSE;
}
 
BOOL CreateFont(MQDocument doc)
{
	int vertex_count, object_index;

	// ダイアログを表示
	if(DialogBox(hInstance, MAKEINTRESOURCE(DLG_FONT), 
			MQ_GetWindowHandle(), FontDialogProc) != TRUE)
		return FALSE;

	// 文字列オブジェクトを作成
	MQObject obj = CreateTextObject(font.text, font.font, font.segment);
	if(obj == NULL)
		return FALSE;

	// サイズを100倍に引き延ばす
	vertex_count = obj->GetVertexCount();
	for(int i=0; i<vertex_count; i++)
		obj->SetVertex(i, obj->GetVertex(i) * 100.f);

	// ドキュメントにオブジェクトを登録し、カレントにする
	object_index = doc->AddObject(obj);
	doc->SetCurrentObjectIndex(object_index);

	return TRUE;
}

